#include "cflironecontrol.h"

//-Конструктор класса--------------------------------------------------------
CFlirOneControl::CFlirOneControl(void)
{
 ImageBuffer=new unsigned char[IMAGE_BUFFER_SIZE];
 ImageBufferSize=0;
 ImageIndex=0;

 for(long n=0;n<256;n++)
 {
  ColorMap_R[n]=(unsigned char)n;
  ColorMap_G[n]=(unsigned char)n;
  ColorMap_B[n]=(unsigned char)n;
 }
 Path=new char[255];
 sprintf(Path,""); 
}
//-Деструктор класса---------------------------------------------------------
CFlirOneControl::~CFlirOneControl()
{
 delete[](ImageBuffer);
 delete[](Path);
}
//-Функции класса------------------------------------------------------------

//----------------------------------------------------------------------------------------------------
//задать рабочий каталог
//----------------------------------------------------------------------------------------------------
void CFlirOneControl::SetPath(char *path)
{
 delete[](Path);
 Path=new char[strlen(path)+1];
 strcpy(Path,path);
}
//----------------------------------------------------------------------------------------------------
//создать изображение
//----------------------------------------------------------------------------------------------------
bool CFlirOneControl::CreateImage(unsigned char *buffer,unsigned long size)
{
 unsigned long n;
 long x;
 long y;
 if (size+ImageBufferSize>=IMAGE_BUFFER_SIZE) ImageBufferSize=0;//начинаем сборку заново
 //копируем данные
 for(n=0;n<size;n++) ImageBuffer[n+ImageBufferSize]=buffer[n];
 ImageBufferSize+=size; 
 if (ImageBufferSize<4) return(false);

 //если в принятых данных встретились 4 специальных байта, то тоже начинаем сборку заново
 unsigned char magicbyte[4]={0xEF,0xBE,0x00,0x00};
 for(n=0;n<ImageBufferSize-4;n++)
 {
  long m;
  bool ok=true;
  for(m=0;m<4;m++)
  {
   if (ImageBuffer[n+m]==magicbyte[m]) continue;
   ok=false;
   break;
  }
  if (ok==true)//байты встретились
  {
   //переносим буфер в начало
   long size=ImageBufferSize-n-4;
   for(m=0;m<size;m++) ImageBuffer[m]=ImageBuffer[m+n+4];
   ImageBufferSize=size;
   n=0;//начинаем сканирование заново
  }
 }
 if (ImageBufferSize<24) return(false);//заголовок не собран
 unsigned long FrameSize=*((unsigned long *)(ImageBuffer+8));
 unsigned long ThermalSize=*((unsigned long *)(ImageBuffer+12));
 unsigned long JpgSize=*((unsigned long *)(ImageBuffer+16));
 unsigned long StatusSize=*((unsigned long *)(ImageBuffer+20));
 if (FrameSize+28>ImageBufferSize) return(false);//кадр ещё не собран 

 //кадр собран
 ImageBufferSize=0;
 unsigned short *raw_thermal_image=new unsigned short[160*120];
 unsigned char *rgb_thermal_image=new unsigned char[160*120*4];
 for(n=0;n<160*120;n++)
 {
  raw_thermal_image[n]=128;
  rgb_thermal_image[n*4+0]=128;
  rgb_thermal_image[n*4+1]=128;
  rgb_thermal_image[n*4+2]=128;
  rgb_thermal_image[n*4+3]=0xff;
 }
 //создаём тепловое изображение (164 числа в строке)
 long min=0x10000;
 long max=0x0;
 unsigned short *d_ptr=(unsigned short*)(ImageBuffer+24);
 bool enabled=true;
 for(y=0;y<120;y++) 
 {  
  unsigned short crc=0;
  unsigned short id;
  unsigned short crc_package;
  //читаем заголовок
  id=*d_ptr;
  id&=0x0FFF;
  CalculateCRC(crc,(id>>8)&0xFF);
  CalculateCRC(crc,id&0xFF);
  d_ptr++;
  crc_package=*d_ptr;  
  CalculateCRC(crc,0);
  CalculateCRC(crc,0);
  d_ptr++;
  for(x=0;x<80;x++,d_ptr++)
  {   
   unsigned short value=*d_ptr;
   raw_thermal_image[x+y*160]=value;
   if (value<min) min=value;
   if (value>max) max=value;
   //считаем CRC
   CalculateCRC(crc,*((unsigned char*)(d_ptr)+1));
   CalculateCRC(crc,*((unsigned char*)(d_ptr)));
  }
  if (crc!=crc_package)//CRC не совпало
  {
   enabled=false;
   break;
  }

  crc=0;
  //читаем заголовок
  id=*d_ptr;
  id&=0x0FFF;
  CalculateCRC(crc,(id>>8)&0xFF);
  CalculateCRC(crc,id&0xFF);
  d_ptr++;
  crc_package=*d_ptr;  
  CalculateCRC(crc,0);
  CalculateCRC(crc,0);
  d_ptr++;
  for(x=80;x<160;x++,d_ptr++)
  {   
   unsigned short value=*d_ptr;
   raw_thermal_image[x+y*160]=value;
   if (value<min) min=value;
   if (value>max) max=value;
   //считаем CRC
   CalculateCRC(crc,*((unsigned char*)(d_ptr)+1));
   CalculateCRC(crc,*((unsigned char*)(d_ptr)));
  }
  if (crc!=crc_package)//CRC не совпало
  {
   enabled=false;
   break;
  }
 }


 if (enabled==true)
 {
  //создаём изображение
  long delta=max-min;
  if (delta==0) delta=1;
  for(x=0;x<160;x++)
  {
   for(y=0;y<120;y++)
   {
    long offset=x+y*160;
    long value=raw_thermal_image[offset];
    value-=min;
    value=(value*255)/delta;
    offset=y+(159-x)*120;
    value&=0xff;
    rgb_thermal_image[offset*4+0]=ColorMap_B[value];
    rgb_thermal_image[offset*4+1]=ColorMap_G[value];
    rgb_thermal_image[offset*4+2]=ColorMap_R[value];
    rgb_thermal_image[offset*4+3]=0xff;
   }
  }
  //сохраняем изображение
  char *filename=new char[strlen(Path)+255];
  sprintf(filename,"%s/image/img%04i.tga",Path,ImageIndex);
  printf("Create image %s\r\n",filename);
  SaveTGA(filename,120,160,rgb_thermal_image);
  ImageIndex++;
 }
 delete[](rgb_thermal_image);
 delete[](raw_thermal_image);
 return(true);
}
//----------------------------------------------------------------------------------------------------
//загрузить карту перекодировки изображения
//----------------------------------------------------------------------------------------------------
bool CFlirOneControl::LoadColorMap(char *filename)
{ 
 FILE *file=fopen(filename,"rb");
 if (file==NULL) return(false);
 unsigned char colormap[768];
 if (fread(colormap,sizeof(unsigned char),768,file)<768)
 {
  fclose(file);
  return(false);
 }
 fclose(file);
 for(long n=0;n<256;n++)
 {
  ColorMap_R[n]=colormap[n*3+0];
  ColorMap_G[n]=colormap[n*3+1];
  ColorMap_B[n]=colormap[n*3+2];
 }
 return(true);
}

//----------------------------------------------------------------------------------------------------
//вычислить crc
//----------------------------------------------------------------------------------------------------
void CFlirOneControl::CalculateCRC(unsigned short &crc,unsigned char byte)
{
 crc^=(byte<<8);
 for(unsigned char n=0;n<8;n++) 
 { 
  if (crc&0x8000) crc=(crc<<1)^0x1021;
             else crc<<=1;
 }
}

//----------------------------------------------------------------------------------------------------
//добавить строку в log-файл
//----------------------------------------------------------------------------------------------------
void CFlirOneControl::AddLog(char *string)
{
 return;  
 FILE *file=fopen("log.txt","ab");
 if (file==NULL) return;
 fprintf(file,"%s",string);
 fclose(file);
}
//-Прочее--------------------------------------------------------------------

