//****************************************************************************************************
//подключаемые библиотеки
//****************************************************************************************************
#include <cstdlib>
#include <iostream>
#include <signal.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/neutrino.h>
#include <inttypes.h>
#include <sys/syspage.h>
#include <pthread.h>
#include <unistd.h>
#include <unix.h>
#include <sys/usbdi.h>
#include <errno.h>

#include "cflironecontrol.h"

//****************************************************************************************************
//макроопределения
//****************************************************************************************************
#define USB_FLIR_VENDOR_ID	0x09cb
#define USB_FLIR_PRODUCT_ID	0x1996

//размер буфера URB
#define URB_BUFFER_SIZE 16384

//максимальное количество точек для одного интерфейса
#define MAX_PIPE 10

//максимальное количество устройств
#define MAX_DEVICE 10

//****************************************************************************************************
//глобальные переменные
//****************************************************************************************************

void *data=NULL;
usbd_connection *connection=NULL;

CFlirOneControl cFlirOneControl;//класс расшифровки данных

struct SFlirOne
{
 usbd_pipe *pipe_0[3];//канал для точки 0
 usbd_urb *urb_0[3];//выделенный URB точки 0

 usbd_pipe *pipe_81;//канал для точки 81
 usbd_urb *urb_81;//выделенный URB точки 81

 usbd_pipe *pipe_83;//канал для точки 83
 usbd_urb *urb_83;//выделенный URB точки 83
 	
 usbd_pipe *pipe_85;//канал для точки 85
 usbd_urb *urb_85;//выделенный URB точки 85
} sFlirOne;

struct SDevice
{
 usbd_device *device;
 usbd_device_instance_t *instance;
 unsigned char InterfaceAddr;//адрес интерфейса
 struct SPipe
 {
  unsigned char Addr;//адрес точки
  usbd_pipe *pipe;//канал для данной точки
  usbd_urb *urb;//выделенный URB
 };
 SPipe sPipe_Array[MAX_PIPE];//список конечных точек
 long PipeAmount;//количество конечных точек
};

volatile SDevice sDevice_Array[MAX_DEVICE];//подключённые устройства
volatile long ConnectedDeviceAmount=0;//количество подключённых устройств
volatile long CurrentConnectedDeviceAmount=0;//количество устройств в настоящий момент времени

//****************************************************************************************************
//прототипы функций
//****************************************************************************************************

void Signal_Exit(int s_code);//уничтожение процесса

//****************************************************************************************************
//функции
//****************************************************************************************************

//----------------------------------------------------------------------------------------------------
//вывод сообщения об ошибке
//----------------------------------------------------------------------------------------------------
void PrintUSBDStatus(long status)
{
 if (status&USBD_STATUS_INPROG) printf("USBD_STATUS_INPROG\r\n"); 
 if (status&USBD_STATUS_CMP) printf("USBD_STATUS_CMP\r\n");  
 if (status&USBD_STATUS_CMP_ERR) printf("USBD_STATUS_CMP_ERR\r\n");  
 if (status&USBD_STATUS_TIMEOUT) printf("USBD_STATUS_TIMEOUT\r\n");  
 if (status&USBD_STATUS_ABORTED) printf("USBD_STATUS_ABORTED\r\n"); 
 if (status&USBD_STATUS_CRC_ERR) printf("USBD_STATUS_CRC_ERR\r\n"); 
 if (status&USBD_STATUS_BITSTUFFING) printf("USBD_STATUS_BITSTUFFING\r\n"); 
 if (status&USBD_STATUS_TOGGLE_MISMATCH) printf("USBD_STATUS_TOGGLE_MISMATCH\r\n"); 
 if (status&USBD_STATUS_STALL) printf("USBD_STATUS_STALL\r\n"); 
 if (status&USBD_STATUS_DEV_NOANSWER) printf("USBD_STATUS_DEV_NOANSWER\r\n"); 
 if (status&USBD_STATUS_PID_FAILURE) printf("USBD_STATUS_PID_FAILURE\r\n"); 
 if (status&USBD_STATUS_BAD_PID) printf("USBD_STATUS_BAD_PID\r\n"); 
 if (status&USBD_STATUS_DATA_OVERRUN) printf("USBD_STATUS_DATA_OVERRUN\r\n"); 
 if (status&USBD_STATUS_DATA_UNDERRUN) printf("USBD_STATUS_DATA_UNDERRUN\r\n"); 
 if (status&USBD_STATUS_BUFFER_OVERRUN) printf("USBD_STATUS_BUFFER_OVERRUN\r\n"); 
 if (status&USBD_STATUS_BUFFER_UNDERRUN) printf("USBD_STATUS_BUFFER_UNDERRUN\r\n"); 
 if (status&USBD_STATUS_NOT_ACCESSED) printf("USBD_STATUS_NOT_ACCESSED\r\n");     
}

//----------------------------------------------------------------------------------------------------
//удалить все ресурсы, связанные с ранее подключёнными устройствами
//----------------------------------------------------------------------------------------------------
void ReleaseConnectedDevices(void)
{
 printf("Release connected devices.\r\n");
 for(long n=0;n<ConnectedDeviceAmount;n++)
 {
  printf("\tIndex: 0x%02x\r\n",n);
  volatile SDevice *sDevice_Ptr=&(sDevice_Array[n]);
  //удаляем данные для конечных точек  
  for(long m=0;m<sDevice_Ptr->PipeAmount;m++)
  {
   usbd_close_pipe(sDevice_Ptr->sPipe_Array[m].pipe);//закрываем канал
   //удаляем память URB
   if (sDevice_Ptr->sPipe_Array[m].urb!=NULL) usbd_free_urb(sDevice_Ptr->sPipe_Array[m].urb);
  } 
  sDevice_Ptr->PipeAmount=0;
  //отключаем устройство	
  usbd_detach(sDevice_Ptr->device);
 }
 ConnectedDeviceAmount=0;
}


//----------------------------------------------------------------------------------------------------
//устройство подключено
//----------------------------------------------------------------------------------------------------
void InsertionFlirOne(usbd_connection *connect,usbd_device_instance_t *instance)
{
 printf("Interfaces of Flir One Gen 2 is connected.\r\n");
 SDevice sDevice;
 sDevice.instance=instance;
 sDevice.device=NULL;
 sDevice.PipeAmount=0;
 if (usbd_attach(connect,sDevice.instance,0,&(sDevice.device))!=EOK)//сбой
 {
  printf("\t\tusbd_attach error!\r\n");
  return;
 }
 printf("\t\tAttach device is ok.\r\n");
 //настраиваем конфигурацию
 //if (usbd_select_config(sDevice.device,3)!=EOK) printf("\t\tusbd_select_config error!\r\n");//третья конфигурация 
 //открываем каналы для конечных точек
 usbd_descriptors_t *desc;
 usbd_desc_node *ifc,*ept;
 //получаем дескриптор интерфейса подключённого устройства
 usbd_interface_descriptor_t *interface;
 interface=usbd_interface_descriptor(sDevice.device,sDevice.instance->config,sDevice.instance->iface,sDevice.instance->alternate,&ifc);
 if (interface!=NULL)//интерфейс получен
 {
  sDevice.InterfaceAddr=interface->bInterfaceNumber;
  printf("\t\tInterface 0x%02x\r\n",sDevice.InterfaceAddr);
  //идём по всем конечным точкам устройста 	
  for(long eix=0;(desc=usbd_parse_descriptors(sDevice.device,ifc,USB_DESC_ENDPOINT,eix,&ept))!=NULL;eix++)
  {
   int addr=desc->endpoint.bEndpointAddress;  	
   if (desc->endpoint.bmAttributes==USB_ATTRIB_CONTROL) printf("\t\t\tEndpoint 0x%02x is Control.\r\n",addr);
   if (desc->endpoint.bmAttributes==USB_ATTRIB_ISOCHRONOUS) printf("\t\t\tEndpoint 0x%02x is Isochronus.\r\n",addr);
   if (desc->endpoint.bmAttributes==USB_ATTRIB_BULK) printf("\t\t\tEndpoint 0x%02x is bulk.\r\n",addr);
   if (desc->endpoint.bmAttributes==USB_ATTRIB_INTERRUPT) printf("\t\t\tEndpoint 0x%02x is interrupt.\r\n",addr);
   //создаём каналы с этими конечными точками
   SDevice::SPipe sPipe;
   sPipe.Addr=addr;   
   if (usbd_open_pipe(sDevice.device,desc,&sPipe.pipe)!=EOK)
   {
    printf("\t\t\tError open pipe for endpoint!\r\n");
    continue;
   }
   else
   {
    //выделяем URB для конечной точки
    sPipe.urb=usbd_alloc_urb(NULL);   	
    if (sPipe.urb==NULL) printf("\t\t\tusbd_alloc_urb error!\r\n");
    if (sDevice.PipeAmount<MAX_PIPE)
    {
     sDevice.sPipe_Array[sDevice.PipeAmount]=sPipe;
     sDevice.PipeAmount++;
    }
    printf("\t\t\tOpen pipe is ok.\r\n");
   }
  }
 } 
 else printf("\t\tGet interface descriptor error!\r\n");
 if (ConnectedDeviceAmount<MAX_DEVICE)
 {
  ((SDevice*)sDevice_Array)[ConnectedDeviceAmount]=sDevice;//добавляем устройство
  ConnectedDeviceAmount++;
 }
 CurrentConnectedDeviceAmount++;
}
 

//----------------------------------------------------------------------------------------------------
//устройство отключено
//----------------------------------------------------------------------------------------------------
void RemovalFlirOne(usbd_connection *connect,usbd_device_instance_t *instance)
{
 printf("Interfaces of Flir One Gen 2 is disconnected.\r\n");
 if (CurrentConnectedDeviceAmount>0) CurrentConnectedDeviceAmount--;
}

//----------------------------------------------------------------------------------------------------
//принимаются данные
//----------------------------------------------------------------------------------------------------
void DataReceive_85(usbd_urb *urb,usbd_pipe *pipe,void *data)
{ 	
 _uint32 status;
 _uint32 length; 
 int error=usbd_urb_status(sFlirOne.urb_85,&status,&length);
 if (error!=EOK) return;
 cFlirOneControl.CreateImage((unsigned char*)data,length);
}

//----------------------------------------------------------------------------------------------------
//основная функция программы
//----------------------------------------------------------------------------------------------------
int main(int argc,char *argv[]) 
{
 //устанавливаем приоритет	
 pid_t pid=getpid();
 setprio(pid,35);
 //настроим системный тик
 struct _clockperiod new_clock;
 new_clock.nsec=100000;//0.1 мс
 new_clock.fract=0;
 ClockPeriod_r(CLOCK_REALTIME,&new_clock,NULL,0);
 	
 //создаём директорию
 mkdir("./image",S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH |S_IXOTH);	
 //загружаем палитры
 cFlirOneControl.LoadColorMap("./iron.pal");
 //устанавливаем рабочий каталог
 cFlirOneControl.SetPath("./");
	
 int error;	
 _uint32 status;
 _uint32 length; 
 //подключим сигналы
 signal(SIGINT,Signal_Exit);
 signal(SIGTERM,Signal_Exit);
 signal(SIGKILL,Signal_Exit);
 signal(SIGCLD,Signal_Exit);
 signal(SIGQUIT,Signal_Exit); 
		
 usbd_device_ident_t interest= 
 {
  USB_FLIR_VENDOR_ID,
  USB_FLIR_PRODUCT_ID,
  USBD_CONNECT_WILDCARD,
  USBD_CONNECT_WILDCARD,
  USBD_CONNECT_WILDCARD,
 }; 

 usbd_funcs_t funcs= 
 {
  _USBDI_NFUNCS,
  &InsertionFlirOne,
  &RemovalFlirOne,
  NULL
 };

 usbd_connect_parm_t cparms= 
 {
  NULL,
  USB_VERSION,
  USBD_VERSION,
  0,
  argc, 
  argv,
  0,
  &interest,
  &funcs
 };	

 //соединяемся со стеком USB	
 error=usbd_connect(&cparms,&connection);	
 if (error!=EOK)
 { 	
  printf("usbd_connect error %i!\r\n",error);
  if (error==EPROGMISMATCH) printf("EPROGMISMATCH\r\n");
  if (error==ENOMEM) printf("ENOMEM\r\n");
  if (error==ESRCH) printf("ESRCH\r\n");
  if (error==EACCES) printf("EACCES\r\n");
  if (error==EAGAIN) printf("EAGAIN\r\n");
  exit(EXIT_FAILURE);
 }
 data=usbd_alloc(URB_BUFFER_SIZE);
 if (data==NULL)
 {
  printf("usbd_alloc error!\r\n");
  //отсоединяемся от стека USB
  usbd_disconnect(connection);
  exit(EXIT_FAILURE);
 }
 while(1)
 {
  //ждём готовности трёх интерфейсов
  if (CurrentConnectedDeviceAmount!=3)
  {
   delay(10);
   continue;	
  }
  //подключены все три устройства Flir One Gen 2  
  //определим каналы точек
  sFlirOne.pipe_0[0]=NULL;
  sFlirOne.urb_0[0]=NULL;
  sFlirOne.pipe_0[1]=NULL;
  sFlirOne.urb_0[1]=NULL;
  sFlirOne.pipe_0[2]=NULL;
  sFlirOne.urb_0[2]=NULL;
  
  sFlirOne.pipe_81=NULL;
  sFlirOne.urb_81=NULL;
  sFlirOne.pipe_83=NULL;
  sFlirOne.urb_83=NULL;
  sFlirOne.pipe_85=NULL;
  sFlirOne.urb_85=NULL; 


  for(long n=0;n<ConnectedDeviceAmount;n++)
  {
   volatile SDevice *sDevice_Ptr=&(sDevice_Array[n]);   
   for(long m=0;m<sDevice_Ptr->PipeAmount;m++)
   {
    volatile SDevice::SPipe *sPipe_Ptr=&(sDevice_Ptr->sPipe_Array[m]);
    if (sPipe_Ptr->Addr==0x85)
    {
     sFlirOne.pipe_85=sPipe_Ptr->pipe;
     sFlirOne.urb_85=sPipe_Ptr->urb;
     printf("Found endpoint 0x85.\r\n");
    } 
    if (sPipe_Ptr->Addr==0x83)
    {
     sFlirOne.pipe_83=sPipe_Ptr->pipe;
     sFlirOne.urb_83=sPipe_Ptr->urb;
     printf("Found endpoint 0x83.\r\n");
    } 
    if (sPipe_Ptr->Addr==0x81)
    {
     sFlirOne.pipe_81=sPipe_Ptr->pipe;
     sFlirOne.urb_81=sPipe_Ptr->urb;
     printf("Found endpoint 0x81.\r\n");
    }
    if (sPipe_Ptr->Addr==0x00)
    {
     long i=sDevice_Ptr->InterfaceAddr;	
     sFlirOne.pipe_0[i]=sPipe_Ptr->pipe;
     sFlirOne.urb_0[i]=sPipe_Ptr->urb;
     printf("Found endpoint 0x00 interface 0x%02x.\r\n",i);
    }
   }
  }  
  //передаём команду на остановку интерфейса 1
  for(long n=0;n<3;n++)
  {
   error=usbd_setup_vendor(sFlirOne.urb_0[n],URB_DIR_OUT,USB_SET_INTERFACE,USB_RECIPIENT_INTERFACE|USB_TYPE_STANDARD,0,1,data,8);
   if (error!=EOK) printf("usbd_setup_vendor error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_0[n],sFlirOne.pipe_0[n],NULL,data,100);
   if (error!=EOK) printf("usbd_io error %i!\r\n",error);
   error=usbd_urb_status(sFlirOne.urb_0[n],&status,&length);
   if (error!=EOK) printf("usbd_urb_status error %i!",error);
   else
   {
    printf("usbd_urb_status return: %ld and length: %ld\r\n",(long)status,(long)length);
    PrintUSBDStatus((long)status);
   }
  } 
  //передаём команду на остановку интерфейса 2
  for(long n=0;n<3;n++)
  {
   error=usbd_setup_vendor(sFlirOne.urb_0[n],URB_DIR_OUT|URB_SHORT_XFER_OK,USB_SET_INTERFACE,USB_RECIPIENT_INTERFACE|USB_TYPE_STANDARD,0,2,data,8);
   if (error!=EOK) printf("usbd_setup_vendor error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_0[n],sFlirOne.pipe_0[n],NULL,data,200);
   if (error!=EOK) printf("usbd_io error %i!\r\n",error);
   error=usbd_urb_status(sFlirOne.urb_0[n],&status,&length);
   if (error!=EOK) printf("usbd_urb_status error %i!",error);
   else
   {
    printf("usbd_urb_status return: %ld and length: %ld\r\n",(long)status,(long)length);
    PrintUSBDStatus(status);
   }
  } 
  //передаём команду на включение интерфейса 1
  for(long n=0;n<3;n++)
  {
   error=usbd_setup_vendor(sFlirOne.urb_0[n],URB_DIR_OUT|URB_SHORT_XFER_OK,USB_SET_INTERFACE,USB_RECIPIENT_INTERFACE|USB_TYPE_STANDARD,1,1,data,8);
   if (error!=EOK) printf("usbd_setup_vendor error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_0[n],sFlirOne.pipe_0[n],NULL,data,200);
   if (error!=EOK) printf("usbd_io error %i!\r\n",error);
   error=usbd_urb_status(sFlirOne.urb_0[n],&status,&length);
   if (error!=EOK) printf("usbd_urb_status error %i!",error);
   else
   {
    printf("usbd_urb_status return: %ld and length: %ld\r\n",(long)status,(long)length);
    PrintUSBDStatus(status);
   }
  } 
  { 
   error=usbd_setup_bulk(sFlirOne.urb_81,URB_DIR_IN,data,URB_BUFFER_SIZE);
   if (error!=EOK) printf("usbd_setup_bulk 0x81 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_81,sFlirOne.pipe_81,NULL,data,10);
   if (error!=EOK) printf("usbd_io 0x81 error %i!\r\n",error);
   
   error=usbd_setup_bulk(sFlirOne.urb_83,URB_DIR_IN,data,URB_BUFFER_SIZE);
   if (error!=EOK) printf("usbd_setup_bulk 0x83 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_83,sFlirOne.pipe_83,NULL,data,10);
   if (error!=EOK) printf("usbd_io 0x83 error %i!\r\n",error);
  }
  //передаём команду на включение интерфейса 2
  for(long n=0;n<3;n++)
  {
   error=usbd_setup_vendor(sFlirOne.urb_0[n],URB_DIR_OUT|URB_SHORT_XFER_OK,USB_SET_INTERFACE,USB_RECIPIENT_INTERFACE|USB_TYPE_STANDARD,1,2,data,8);
   if (error!=EOK) printf("usbd_setup_vendor error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_0[n],sFlirOne.pipe_0[n],NULL,data,200);
   if (error!=EOK) printf("usbd_io error %i!\r\n",error);
   error=usbd_urb_status(sFlirOne.urb_0[n],&status,&length);
   if (error!=EOK) printf("usbd_urb_status error %i!",error);
   else
   {
    printf("usbd_urb_status return: %ld and length: %ld\r\n",(long)status,(long)length);
    PrintUSBDStatus(status);
   }
  }
  { 
   error=usbd_setup_bulk(sFlirOne.urb_81,URB_DIR_IN,data,URB_BUFFER_SIZE);
   if (error!=EOK) printf("usbd_setup_bulk 0x81 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_81,sFlirOne.pipe_81,NULL,data,10);
   if (error!=EOK) printf("usbd_io 0x81 error %i!\r\n",error);
   
   error=usbd_setup_bulk(sFlirOne.urb_83,URB_DIR_IN,data,URB_BUFFER_SIZE);
   if (error!=EOK) printf("usbd_setup_bulk 0x83 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_83,sFlirOne.pipe_83,NULL,data,10);
   if (error!=EOK) printf("usbd_io 0x83 error %i!\r\n",error);
  }
  //принимаем данные
  while(1)
  {
   delay(1);
   if (CurrentConnectedDeviceAmount!=3) break;//устройство отключено
   error=usbd_setup_bulk(sFlirOne.urb_85,URB_DIR_IN|URB_SHORT_XFER_OK,data,URB_BUFFER_SIZE);
   if (error!=EOK) printf("usbd_setup_bulk 0x85 error %i\r\n",error);
   error=usbd_io(sFlirOne.urb_85,sFlirOne.pipe_85,DataReceive_85,data,0);
   if (error==ENODEV) break;
   /*
   if (error!=EOK) printf("usbd_io 0x85 error %i\r\n",error);
   else
   {
    error=usbd_urb_status(sFlirOne.urb_85,&status,&length);
    if (error!=EOK && error!=EBUSY) printf("usbd_urb_status 0x85 error %i!\r\n",error);
    else
    {
     printf("usbd_urb_status return: %i and length: %i\r\n",(long)status,(long)length);
     PrintUSBDStatus(status);
    }
   }*/
   error=usbd_setup_bulk(sFlirOne.urb_81,URB_DIR_IN|URB_SHORT_XFER_OK,data,URB_BUFFER_SIZE);
   //if (error!=EOK) printf("usbd_setup_bulk 0x81 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_81,sFlirOne.pipe_81,NULL,data,0);
   //if (error!=EOK) printf("usbd_io 0x81 error %i!\r\n",error);
   
   error=usbd_setup_bulk(sFlirOne.urb_83,URB_DIR_IN|URB_SHORT_XFER_OK,data,URB_BUFFER_SIZE);
   //if (error!=EOK) printf("usbd_setup_bulk 0x83 error %i!\r\n",error);
   error=usbd_io(sFlirOne.urb_83,sFlirOne.pipe_83,NULL,data,0);
   //if (error!=EOK) printf("usbd_io 0x83 error %i!\r\n",error);
  } 
  //ждём отключения всех устройств
  while(1)
  {
   delay(10);
   if (CurrentConnectedDeviceAmount!=0) continue;
   break;
  }
  //освобождаем все занятые ресурсы
  ReleaseConnectedDevices();
  delay(1000);
 }
 usbd_free(data);
 //отсоединяемся от стека USB
 usbd_disconnect(connection);
 return(EXIT_SUCCESS);
}

//----------------------------------------------------------------------------------------------------
//сигналы
//----------------------------------------------------------------------------------------------------
//уничтожение подчинённого процесса
void Signal_Exit(int s_code)
{
 ReleaseConnectedDevices(); 
 if (data!=NULL) usbd_free(data); 
 if (connection!=NULL) usbd_disconnect(connection); 
 printf("Flir One exit.\n");
 exit(EXIT_SUCCESS);
}
